require 'sass/scss/rx'
require 'sass/scss/parser'
require 'sass/scss/static_parser'
require 'sass/scss/css_parser'

module Sass
  # SCSS is the CSS syntax for Sass.
  # It parses into the same syntax tree as Sass,
  # and generates the same sort of output CSS.
  #
  # This module contains code for the parsing of SCSS.
  # The evaluation is handled by the broader {Sass} module.
  module SCSS; end
end
